/*
 * Author: Jake Goodwin
 * Date: 2023
 * Description: Small library for using standard rotary encoders.
 */

/*
 * Rotary pin state bits
 * | NOUSE |A0 |B0 |A1 |B1 |
 * | 0000b | 0 | 0 | 0 | 0 |
 */

/* Rotary encoder mechanical contacts.
 * Clockwise sequence:
 *      | A B | DEC
 * t1   | 0 1 | 0x1
 * t2   | 0 0 | 0x0
 * t3   | 1 0 | 0x2
 * t4   | 1 1 | 0x3
 *
 * Counter-Clockwise sequence:
 *      | A B | Hex
 * t1   | 1 0 | 0x2
 * t2   | 0 0 | 0x0
 * t3   | 0 1 | 0x1
 * t4   | 1 1 | 0x3
 *
 * If tossed into a 8bit register or var:
 * CW:  0b01001011 = 75
 * CCW: 0b10000111 = 135
 */


#include <avr/io.h>
#include <stdint.h>
#include <util/delay.h>

// PIN DEFINITIONS
//PUT YOUR PINS IN FOR ENCODER A/B
//Along with pin offset if needed. So i'm using pins 1,2 so because I 
//skipped 0 I set the offset too 1.

#define ENCODER_A       PINC1
#define ENCODER_B       PINC2
#define PIN_OFFSET      1
#define ENCODER_DDR     DDRC
#define ENCODER_PORT    PORTC

#define INPUT_REG       PINC

#define LED_PIN     PINC0

//#############################
//Globals
//#############################

//if you want you can replace this with just encoder_byte
//and bitshift/mask to compare it with the pins state, but it's less readable.
uint8_t old_state = 0x0;
uint8_t encoder_byte = 0x0;

//A more friendly way to check state.
enum encoder_dir {
    CW = 1,
    CCW = 2,
    none = 0,
};


//#############################
//FUNCTIONS
//#############################
/*
 * Input: None
 * Output: None
 * Description: Toggles the pin for the LED indicator.
 */
void led_blink(uint8_t x) {
    //Set the DDR for output.
    DDRC |= (1<<LED_PIN);

    for(uint8_t i = 0; i < x; i++) {
        PORTC ^= (1<<LED_PIN);
        _delay_ms(500);
        PORTC ^= (1<<LED_PIN);
        _delay_ms(500);
    }
    
}



void init_encoder(void) {

    //Then encoder has two outputs, A, B also labeled as clk, dt.
    //Set the two pins too monitor digital inputs
    ENCODER_DDR &= ~( (1<<ENCODER_A)|(1<<ENCODER_B) );
    ENCODER_PORT &= ~( (1<<ENCODER_A)|(1<<ENCODER_B) );
}

/*
 * Input: None,
 * Output: rotary_state: 0=No change, 1= Clockwise tick, 2= counter clockwise tick.
 * Description: checks the rotary encoder for changes in it's state. It updates 
 * the global encoder byte by taking the state of A and B pins from the encoder
 * and shifting their state bits into it 2bits per shift. After four shifts 
 * the value in the byte should equal 75 or 135 which would mean the encoder
 * completed a single step or tick.
 */ 
uint8_t poll_encoder() {
    //in this scope we have defined the state this way
    enum state {
        CW = 75,
        CCW = 135,
    };

    //clear out old state bits.
    uint8_t encoder_state = 0x0;

    //mask out the other pins, and deal with the pin offset.
    encoder_state |= ( (INPUT_REG & ((1<<ENCODER_A)|(1<<ENCODER_B))  )>> PIN_OFFSET);

    //we want to check for the correct sequence of bits.
    //we also only want to accept new states as input. Not old states.
    if(encoder_state != old_state){
        old_state = encoder_state;
        encoder_byte = (encoder_byte<<2);
        encoder_byte |= encoder_state;
    }

    switch(encoder_byte) {
        case CCW:
            encoder_byte = 0x0;
            encoder_state= 2;
            break;
        case CW:
            encoder_byte = 0x0;
            encoder_state= 1;
            break;
        default:
            encoder_state= 0x0;
            break;
    }

    return encoder_state;

}




//This is just an example function of how you could use it.
int demo() {
    init_encoder();
    uint8_t volume = 0;
    led_blink(2);

    while(1) {
        switch(poll_encoder()) {
            case CW:
                if(volume < 100){volume++;}
                led_blink(1);
                break;
            case CCW:
                if(volume > 0){volume--;}
                led_blink(2);
                break;
            default:
                ;//nothing.
        }
        
        //Do other polling operations here
    }
    return 0;
}


